/*
 * Copyright (C) 2013 by the University of Southern California
 * $Id$
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include "utility.hh"
#include "anon.hh"
#include "aes_util.hh"
#include "base64_util.hh"
#include "cryptopANT.h"

//#define USE_HMAC
//#define USE_AES

#ifdef USE_HMAC
#include <openssl/hmac.h> 
#endif //USE_HMAC

#include <openssl/evp.h>
#include <openssl/bio.h>
#include <openssl/buffer.h>

#include <ldns/ldns.h>
#define SHA1_DIGEST_LENGTH 20

#define KEEP 		0
#define DO_NAME 	1
#define DO_IP		2
#define DO_ALL 	3
#define ZERO_ITEM 	4
#define ENCRYPT 	5

using namespace std;

/*
	split a string
*/
void
str_split2(const string &s, const char *spt, vector<string> &substrs)
{
	const string delimiters = spt;
	size_t current = 0;
	size_t next = 0;
	do {
		next = s.find( delimiters, current );
		substrs.push_back(s.substr(current, next - current));
		current = next + delimiters.size();
	} while (next != string::npos);
}
 
/*
	get the ip address; anonymize it if needed
*/
void
get_ip_str(struct sockaddr *ip, string &ipstr, bool anon, int pbits)
{
	int pass_bits4 = pbits & 0x0000ffff;
	int pass_bits6 = (pbits & 0xffff0000) >> 16;
	if (ip->sa_family == AF_INET) {	//ipv4		
		char ipv4str[INET_ADDRSTRLEN];
		memset(ipv4str, '\0', INET_ADDRSTRLEN);
		
		struct sockaddr_in *ipv4 = (struct sockaddr_in *)ip;
		inet_ntop(AF_INET, &(ipv4->sin_addr), ipv4str, INET_ADDRSTRLEN);
		ipstr = ipv4str;
#ifdef DEBUG	
		cout<<"1->ipv4[Before anon]The incoming IP is: "<<ipstr<<endl;	
#endif		
		if (anon) {
			uint32_t ipaddr = 0;
			ipaddr = ipv4->sin_addr.s_addr;
			
			char anonip[INET_ADDRSTRLEN];
			memset(anonip, '\0', INET_ADDRSTRLEN);
			
			struct sockaddr_in sa;
			sa.sin_addr.s_addr = scramble_ip4(ipaddr, pass_bits4);
			
			inet_ntop(AF_INET, &(sa.sin_addr), anonip, INET_ADDRSTRLEN);
			ipstr = anonip;
#ifdef DEBUG
			cout<<"2->ipv4[After anon]The incoming IP is: "<<anonip<<endl;
			struct sockaddr_in sa_tmp;
			sa_tmp.sin_addr.s_addr = unscramble_ip4(sa.sin_addr.s_addr, pass_bits4);
			cout<<"pass bits:"<<pass_bits4<<endl;
			inet_ntop(AF_INET, &(sa_tmp.sin_addr), anonip, INET_ADDRSTRLEN);
			//ipstr = anonip;
			cout<<"3->ipv4[recovery]The incoming IP is: "<<anonip<<endl;
#endif					
		}
	} 	
	
	if (ip->sa_family == AF_INET6) {
		char ipv6str[INET6_ADDRSTRLEN];
		memset(ipv6str, '\0', INET6_ADDRSTRLEN);
		
		struct sockaddr_in6 *ipv6 = (struct sockaddr_in6 *)ip;
		inet_ntop(AF_INET6, &(ipv6->sin6_addr), ipv6str, INET6_ADDRSTRLEN);
		ipstr = ipv6str;
#ifdef DEBUG	
		cout<<"ipv6[Before anon]The incoming IP is: "<<ipstr<<endl;
		cout<<"pass bits:"<<pass_bits6<<endl;
#endif		
		if (anon) {
			scramble_ip6(&(ipv6->sin6_addr), pass_bits6);
			
			char anonip[INET6_ADDRSTRLEN];
			memset(anonip, '\0', INET6_ADDRSTRLEN);

			inet_ntop(AF_INET6, &(ipv6->sin6_addr), anonip, INET6_ADDRSTRLEN);
			ipstr = anonip;
#ifdef DEBUG
			cout<<"ipv6[After anon]The incoming IP is: "<<anonip<<endl;
#endif
		}
	}
	if (anon)
		ipstr += "*";
		
}

/*
	remove the newline in a base64 string
*/
/*void
bash64_remove_newline(string &s)
{
	istringstream ss(s);
	s = "";
	while(!ss.eof()) {
		string x;
		ss >> x;
		s += x;
	}
}*/

/*
	anon label
*/
void
anon_string(char *s, unsigned char *key, size_t keylen, 
		unsigned char *iv, size_t ivlen, string &result, bool reverse_mode)
{
	unsigned char *aes_str = NULL;
	int len = -1;
	char *b64_str = NULL;
	
	/*
		make a copy of the key and iv, better not touch the original ones, delete the copy in the end
	*/
	unsigned char *key_copy = NULL;
	unsigned char *iv_copy = NULL;	
	key_copy = new unsigned char[keylen];
	iv_copy = new unsigned char[ivlen];
	if (key_copy == NULL || iv_copy == NULL) {
		cerr << "[ERROR] memory allocation problem: key_copy == NULL || iv_copy == NULL)" << endl;
		exit(-1);
	}
	memcpy (key_copy, key, keylen);
	memcpy (iv_copy, iv, ivlen);
#ifdef USE_HMAC
	if (reverse_mode) {
		cerr << "[ERROR] reverse_mode is not supported in hash mac" << endl;
		exit(-1);
	}
	HMAC_CTX      ctx;
	unsigned int  len;
	out = new unsigned char[SHA1_DIGEST_LENGTH];		//store hash result
	out_len = SHA1_DIGEST_LENGTH;
	HMAC_Init(&ctx, key_copy, keylen, EVP_sha1());		//sha1
	HMAC_Update(&ctx, (const unsigned char*)s, strlen(s));
	HMAC_Final(&ctx, out, &len);				//get hash
	HMAC_cleanup(&ctx);					//clean up

#else 	//USE_AES
	//int len = strlen(s);
	//len += 1;//include the '\0'
	if (reverse_mode) {//base64_decode => decrypt string
		len = strlen(s);
		b64_str = base64_decode((unsigned char*)s, len);
		//cout<<"strlen(b64_str)="<<len<<endl;
		aes_str = aec_decrypt((unsigned char *)b64_str, len, key_copy, iv_copy);
		if (aes_str == NULL) {
			cerr << "[ERROR] aec_decrypt return NULL!" << endl;
			exit(-1);
		}
		result = (char *)aes_str;
		//cout<<"result="<<result<<endl;
		delete[] aes_str;
		delete[] b64_str;
	} else {//encrypt => base64_encode
		len = strlen(s);
		len += 1;//include the '\0'
		aes_str = aec_encrypt((unsigned char *)s, len, key_copy, iv_copy);
		if (aes_str == NULL) {
			cerr << "[ERROR] aec_encrypt return NULL!" << endl;
			exit(-1);
		}
		b64_str = base64_encode(aes_str, len);
		result = b64_str;
		delete[] aes_str;
		delete[] b64_str;
	}
	
#endif	//USE_HMAC

	if (key_copy != NULL)
		delete[] key_copy;
	if (iv_copy != NULL)
		delete[] iv_copy;	
}

/*
	anon domain name
*/
void
anon_name(char *query, unsigned char *key, size_t keylen, 
	unsigned char *iv, size_t ivlen, string &result, bool reverse_mode)
{
	if (query == NULL) {	//empty name, return empty
		result = "NULL-EMPTY";
		return;
	} else {
		result = "";
	}
	string qname = query;
	const string DOT = ".";
	
	if (qname == DOT) {
		result = qname;
		return;
	}

	string s;
	string sub_result;

	size_t current = 0;
	size_t next = 0;
	//qname="this is only test";
	while (next != string::npos) {
		next = qname.find(DOT, current);
		
		//next can be string::npos here. but for .substr(), if out of string 
		//content range only those to the end of the string will be keept
		s = qname.substr(current, next - current);
		//cout<<"s="<<s<<endl;
		if (s.size() > 0) {//skip empty string
			anon_string((char *)s.c_str(), key, keylen, iv, ivlen, sub_result, reverse_mode);
			result += sub_result;	
		}
		if (next != string::npos) result += DOT;
		current = next + DOT.size();	
	}
	//cout<<"result="<<result<<endl;
}

/*
	anonymize ip string
*/
void
anon_ip_str (string &ip_str, unsigned char *key, size_t len, int pbits)
{
	struct sockaddr *ip = NULL;
	struct sockaddr_in sa4;
	struct sockaddr_in6 sa6;
	int resv4 = inet_pton(AF_INET, (const char *)ip_str.c_str(), &(sa4.sin_addr));
	int resv6 = inet_pton(AF_INET6, (const char *)ip_str.c_str(), &(sa6.sin6_addr));	
	if (resv4 == 1) {		//it's an ipv4 address
		sa4.sin_family = AF_INET;
		ip = (sockaddr *)&sa4;
	} else if (resv6 == 1) {	//ipv6
		sa6.sin6_family = AF_INET6;
		ip = (sockaddr *)&sa6;
	}
	
	if (ip != NULL) {
		get_ip_str(ip, ip_str, true, pbits); 
	} 
}

/*
	rebuild the rdata from string vector and insert "%20"
*/
void
build_rdata (vector<string> &sv, string &rd)
{
	rd = "";
	for (size_t i=0; i<sv.size(); i++) {
		rd += sv[i];
		if ((i + 1) < sv.size()){ 
			rd += RDATA_SPT;
		} 
	}
}

/*
	which action should be performed sv[idx]?
*/
void
sub_anon_rdata(int idx, vector<string> &sv, int action, 
	unsigned char *key, size_t len, unsigned char *iv, size_t iv_len, int pbits)
{
	string tmp1 = "";
	string tmp2 = "";

	switch (action) { 
		case DO_NAME:
			tmp1 = sv[idx];
			anon_name((char *)tmp1.c_str(), key, len, iv, iv_len, tmp2, false);
			sv[idx] = tmp2;
			break;	
		case DO_IP:
			tmp1 = sv[idx];
			anon_ip_str(tmp1, key, len, pbits);
			sv[idx] = tmp1;
			break;
		case DO_ALL:
			for (size_t i=0; i<sv.size(); i++) {
				tmp1=""; tmp2="";
				tmp1 = sv[i];
				anon_name((char *)tmp1.c_str(), key, len, iv, iv_len, tmp2, false);
				sv[i] = tmp2;
			}
			break;
		case ENCRYPT:
			tmp1 = sv[idx];
			anon_string((char *)tmp1.c_str(), key, len, iv, iv_len, tmp2, false);
			sv[idx] = tmp2;	
			break;			
		case ZERO_ITEM:
			sv[idx] = "0";
			break;			
		default:
			cerr<<"[ERROR] unkown action in sub_anon_rdata()."<<endl;
			exit(-1);
			break;		
	}
}


/*
	anonymize rdata of answer, additional and authority fields 
	basically, we can specify a policy to each component
*/
void
anon_rdata(int type, string &rdata, unsigned char *key, size_t len, 
	unsigned char *iv, size_t iv_len, int pbits)
{
	bool zero = false;
	if (rdata.size() == 0) {
		return;
	}
	
	//get each components of rdata
	vector<string> sv;
	vector<string> sv2;
	
	string sub_res1;//store the temp result;
	string sub_res2;
	
	switch (type) {
		//the first part are only contain number, domain name, ip addresses
		//the numbers are not informative, only anonymize ip and domain
		
		//type		        format
		case LDNS_RR_TYPE_A:	//IPv4		
		case LDNS_RR_TYPE_AAAA:	//IPv6
			anon_ip_str(rdata, key, len, pbits);
			break;
		case LDNS_RR_TYPE_CNAME:	//domain
		case LDNS_RR_TYPE_MB:	//domain
		case LDNS_RR_TYPE_MD:	//domain
		case LDNS_RR_TYPE_MF:	//domain
		case LDNS_RR_TYPE_MG:	//domain
		case LDNS_RR_TYPE_MR:	//domain
		case LDNS_RR_TYPE_NS:	//domain
		case LDNS_RR_TYPE_PTR:	//domain
		case LDNS_RR_TYPE_DNAME:	//domain
			sub_res1 = rdata;
			anon_name((char *)sub_res1.c_str(), key, len, iv, iv_len, rdata, false);
			break;
		case LDNS_RR_TYPE_MX:	//num domain
		case LDNS_RR_TYPE_AFSDB:	//num domain
		case LDNS_RR_TYPE_RT:	//num domain
		case LDNS_RR_TYPE_KX:	//num domain
			sv.clear();
			str_split(rdata, sv);
			sub_anon_rdata(1, sv, DO_NAME, key, len, iv, iv_len, -1);
			build_rdata(sv, rdata);
			break;			
		case LDNS_RR_TYPE_MINFO:	//domain domain
		case LDNS_RR_TYPE_RP:	//domain domain
			sv.clear();
			str_split(rdata, sv);
			sub_anon_rdata(0, sv, DO_NAME, key, len, iv, iv_len, -1);
			sub_anon_rdata(1, sv, DO_NAME, key, len, iv, iv_len, -1);
			build_rdata(sv, rdata);
			break;	
		case LDNS_RR_TYPE_SRV: 	//num num num domain
			sv.clear();
			str_split(rdata, sv);
			sub_anon_rdata(3, sv, DO_NAME, key, len, iv, iv_len, -1);
			build_rdata(sv, rdata);
			break;
		case LDNS_RR_TYPE_PX: 	//num domain domain
			sv.clear();
			str_split(rdata, sv);
			sub_anon_rdata(1, sv, DO_NAME, key, len, iv, iv_len, -1);
			sub_anon_rdata(2, sv, DO_NAME, key, len, iv, iv_len, -1);
			build_rdata(sv, rdata);
			break;	
		case LDNS_RR_TYPE_NAPTR: 	//num num str str str domain
			sv.clear();
			str_split(rdata, sv);
			sub_anon_rdata(2, sv, ENCRYPT, key, len, iv, iv_len, -1);
			sub_anon_rdata(3, sv, ENCRYPT, key, len, iv, iv_len, -1);
			sub_anon_rdata(4, sv, ENCRYPT, key, len, iv, iv_len, -1);
			sub_anon_rdata(5, sv, DO_NAME, key, len, iv, iv_len, -1);
			build_rdata(sv, rdata);
			zero = true;
			break;	
		case LDNS_RR_TYPE_HINFO: 	//str str (CPU OS)
		case LDNS_RR_TYPE_X25: 	//str
		case LDNS_RR_TYPE_ISDN:	//str str
			sv.clear();
			str_split(rdata, sv);
			sub_anon_rdata(-1, sv, DO_ALL, key, len, iv, iv_len, -1);
			build_rdata(sv, rdata);
			break;
		case LDNS_RR_TYPE_WKS:	//obsoleted?
			sv.clear();	//address protocol bitmap
			str_split(rdata, sv);
			sub_anon_rdata(0, sv, DO_IP, key, len, iv, iv_len, pbits);
			build_rdata(sv, rdata);
			break;
		case LDNS_RR_TYPE_A6:	//num ipv6 domain
			sv.clear();
			str_split(rdata, sv);
			sub_anon_rdata(1, sv, DO_IP, key, len, iv, iv_len, pbits);
			sub_anon_rdata(2, sv, DO_NAME, key, len, iv, iv_len, -1);
			build_rdata(sv, rdata);
			break;			
		case LDNS_RR_TYPE_NXT:	//domain bitmap
		case LDNS_RR_TYPE_NSEC:	//domain bitmap
			sv.clear();
			str_split(rdata, sv);
			sub_anon_rdata(0, sv, DO_NAME, key, len, iv, iv_len, -1);
			build_rdata(sv, rdata);
			break;
		case LDNS_RR_TYPE_SPF: 	//TXT, one or more string
		case LDNS_RR_TYPE_TXT: 	//TXT
		case LDNS_RR_TYPE_NULL:	//anything
			zero = true;
			break;
		case LDNS_RR_TYPE_DNSKEY:	//here the digest might be base64 which has spaces in it
		case LDNS_RR_TYPE_DS:
		case LDNS_RR_TYPE_DLV:
			sv.clear(); sv2.clear();
			sv.resize(4);
			str_split(rdata, sv2);
			for (int i=0; i<3; i++) {
				sv[i] = sv2[i];
			}
			sv[3] = "0";
			//sub_anon_rdata(3, sv, ZERO_ITEM, key, len, iv, iv_len, -1);
			build_rdata(sv, rdata);	
			break;
		case LDNS_RR_TYPE_SOA://name name num num num num num
			sv.clear();
			str_split(rdata, sv);
			sub_anon_rdata(0, sv, DO_NAME, key, len, iv, iv_len, -1);
			sub_anon_rdata(1, sv, DO_NAME, key, len, iv, iv_len, -1);
			build_rdata(sv, rdata);
			break;		
		case LDNS_RR_TYPE_RRSIG:
		case LDNS_RR_TYPE_SIG:
		case LDNS_RR_TYPE_NSEC3:
		case LDNS_RR_TYPE_NSEC3PARAM:
			zero = true;
			break;
		
		default:
			zero = true;
			break;
	}
	if (zero) {
		rdata = "-";
		return;
	}
}

