/*
 * Copyright (C) 2013 by the University of Southern California
 * $Id$
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 */
 
#include <openssl/bio.h>
#include <openssl/evp.h>
#include <openssl/buffer.h>
#include <cstring>
#include <iostream>
using namespace std;

/*
	base64 encode, in any cases, users are responsiable to delete returned char*
*/
char *base64_encode(unsigned char *msg, int &msg_len)
{
	BIO *mem, *b64;
	BUF_MEM *ptr;
	
	b64 = BIO_new(BIO_f_base64());
	mem = BIO_new(BIO_s_mem());
	mem = BIO_push(b64, mem);// b64-->mem, mem pointed to b64 the end of the chain
	BIO_set_flags(mem, BIO_FLAGS_BASE64_NO_NL); 

	BIO_write(mem, msg, msg_len);//write to b64
	int rv = BIO_flush(mem);
	if (rv != 1) {
		cerr<<"[ERROR] BIO_flush fails with return code:"<<rv<<endl;
		BIO_free_all(mem);
		exit(-1);
	}
	BIO_get_mem_ptr(mem, &ptr);//get data from mem

	char *buf = new char[ptr->length+1];
	memset(buf, '\0', ptr->length+1);
	memcpy(buf, ptr->data, ptr->length);
	buf[ptr->length] = '\0';
	BIO_free_all(mem);
	msg_len = ptr->length + 1;
	return buf;
}

/*
	base64 decode, users should delete buf in any case
*/
char *base64_decode(unsigned char *msg, int &msg_len)
{
	BIO *b64, *mem;
	int pad_num = 0;
	if (msg[msg_len-1] == '=') {
		pad_num += 1;
		if (msg[msg_len-2] == '=') {
			pad_num += 1;
		}
	}
	int new_len = int(msg_len * 3 / 4) - pad_num; //this it to avoid mem leak, it's ok to use msg_len here since we set the rest to '0'
	char *buf = new char[new_len + 1];
	memset(buf, '\0', new_len + 1);
	b64 = BIO_new(BIO_f_base64());
	mem = BIO_new_mem_buf(msg, msg_len);
	mem = BIO_push(b64, mem);//b64--->mem, mem pointed to b64, the end of the chain
	BIO_set_flags(mem, BIO_FLAGS_BASE64_NO_NL);
	int read_len = BIO_read(mem, buf, msg_len);//read from b64 but data is at mem sink
	if (read_len != new_len) {
		cerr<<"[ERROR] error in BIO_read, new_len="<<new_len<<" read_len="<<read_len<<endl;
		BIO_free_all(mem);
		exit(1);
	}
	//buf[new_len] = '\0';
	msg_len = new_len;	
	BIO_free_all(mem);
	return buf;
}
